==> file4life-1.1.3-source.tgz: file4life-1.1.3/src/F4L/Actions/Files.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Actions.Files (
  addFiles
  , addTestFiles
  , checkedCopyFileCandidates
  , fileCandidates
  , fileId
  , findLocalFileCandidates
  , getCleanDirectoryContents
) where

import F4L
import F4L.Actions.Common
import F4L.Actions.Ops
import F4L.Actions.Storages
import F4L.BufferMonad
import F4L.F4LMonad
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import F4L.Knowledge
import F4L.Op
import F4L.SeenTree as ST hiding ( removeFile )
import F4L.Utils hiding( fileSize )

import Control.DeepSeq
import Control.Lens
import Control.Monad
import Control.Monad.Error( ErrorT(..), throwError )
import Control.Monad.Reader
import Control.Monad.State.Strict
import Data.Either( partitionEithers )
import Data.List( sort )
import qualified Data.Map.Strict as M -- TODO: just use SM
import qualified Data.Map.Strict as SM
import qualified Data.Set as S
import System.FilePath( combine, joinPath, splitDirectories )


fileId :: FSMonad m => FilePath -> ErrorT Error m FileId
-- fileId path = processLengthAndBLFromFile path $ \ l bs -> FileId l ( sha1 bs )
fileId path = actOnLengthAndBLFromFile path act
  where
    act l bs = do
      let s = sha1 bs
      let res = FileId l s
      s `deepseq` return res


getCleanDirectoryContents dir = 
  getDirectoryContents dir >>= return . filter ( not . flip elem [ ".", ".." ] ) >>= return . sort

checkedCopyFile :: DelFSMonad m => FileId -> FilePath -> FilePath -> ErrorT Error m ()
checkedCopyFile ( FileId l h ) source dest = do
  writeRes <- actOnLengthAndBLFromFile source checkedWrite
  case writeRes of
    Left err -> do
      -- TODO: remove dest file on incomplete write
      throwError err
    Right res -> return res
  where
    checkedWrite i bl = runErrorT $ do
      when ( i /= l ) $
        throwError $ "Length mismatch copying " ++ source ++ " to " ++ dest ++ ": should be " ++ show l ++ ", is " ++ show i
      ctx <- writeBLToFileAndDo dest newSha1 ( liftM ( liftM return ) updateSha1' ) bl
      let rh = toSha1 ctx
      when ( rh /= h ) $ do
        removeFile dest
        throwError $ "Hash mismatch copying " ++ source ++ " to " ++ dest ++ ": should be " ++ hashToS h ++ ", is " ++ hashToS rh
      return ()


checkedCopyFileCandidates :: DelFSMonad m => FileId -> [ FilePath ] -> FilePath -> String -> ErrorT Error ( F4LMonad k m ) ()
checkedCopyFileCandidates fid candidates destPath preMessage =
  foldl mplus ( throwError "No candidate for file" ) $ flip map candidates $ \ sourcePath -> do
    outputLineMsg $ preMessage ++ "Copying " ++ sourcePath ++ " to " ++ destPath
    say Chatty $ "Copying " ++ sourcePath ++ " to " ++ destPath
    checkedCopyFile fid sourcePath destPath


findLocalFileCandidates :: ( FSMonad m, BufferMonad m, KnownDataClass k ) 
                           => FileId -> F4LMonad k m [ FilePath ]
findLocalFileCandidates fid = do
  storages <- storagesForEnv
  fd <- use $ st_knowledge . kd_files . to ( SM.findWithDefault noFileData fid  )
  return $ fileCandidates storages fd


fileCandidates :: SM.Map StorageId [ FilePath ] -> FileData -> [ FilePath ]
fileCandidates storages dat =
  [ combine sseen path
  | ( sid, path ) <- map ( mapSnd shortToPath ) $ S.toList ( dat ^. fd_seen )
  , sseen <- M.findWithDefault [] sid storages
  ]


{-
findFileWithSize :: ( FSMonad m, BufferMonad m ) => Int -> [ FilePath ] -> m ( Maybe FilePath )
findFileWithSize _ [] = return Nothing
findFileWithSize i ( x : xs ) = do
  e <- doesFileExist x
  case e of
    True -> do
      es <- runErrorT $ fileSize x
      case es of
        Right s | s == i -> return ( Just x )
        _ -> findFileWithSize i xs
    False ->
      findFileWithSize i xs
-}


addTestFile :: ( BufferMonad m, Knowledge k ) => Int -> F4LMonad k m ()
addTestFile i = do
  let fid = testFid i
  setFid fid
  setPath $ "test-" ++ show i
  addOp SeenMark
  return ()


addTestFiles :: ( BufferMonad m, Knowledge k ) => Int -> F4LMonad k m ()
addTestFiles i = do
  withStorage "-" "/"
  mapM_ addTestFile [ 1 .. i ]
  return ()


-- | Ops are only added on no error
addFile :: ( FSMonad m, BufferMonad m, Knowledge k, KnownDataClass k ) 
           => RichPath -> Bool -> FileId
           -> ErrorT Error ( F4LMonad k m ) ()
addFile rp samePosition fid = do
  let ( sid, sPath, relPath, path ) = ( rp_sid rp, rp_sPath rp, rp_relPath rp, rp_canPath rp )
  say Chatty $ "Adding " ++ sid ++ ":" ++ sPath ++ ":" ++ relPath
  withExif <- lift $ use st_withExif
  exif_props <- if withExif && not samePosition then lift ( exifProperties path ) else return []
  lift $ do
    withStorage sid sPath
    unless samePosition $ do
      setFid fid
      setPath relPath
      -- TODO: only add new properties
      mapM_ (addOp . attrOp) exif_props
      addOp SeenMark




addFiles :: ( FSMonad m, BufferMonad m, Knowledge k, SeenStoragesClass k, KnownDataClass k, Show k ) => Maybe Priority -> [FilePath] -> F4LMonad k m ()
addFiles mpri paths = do
  st_storages .= SM.empty
  ( withError, richPaths ) <- liftM partitionEithers $ forM paths $ \ p -> do
    res <- runErrorT $ forceGetStorage p
    case res of
      Right res -> return $ Right res
      Left err -> return $ Left ( p, err )
  forM_ withError $ \ ( p, err ) -> do
    say Quiet $ "Warning: error setting StorageId for " ++ p ++ ": " ++ err
  let sidLocations = SM.fromList [ ( rp_sid x, rp_sPath x ) | x <- richPaths ]
  say SayAll $ "richPaths: " ++ show richPaths
  let rootSidPaths = dominants sidPathCmp dominates $ map toSidPath richPaths
  say SayAll $ "rootSidPaths: " ++ show rootSidPaths
  withRescan <- use st_withRescan
  -- accumulate files seen in same position with some size
  say Chatty "Adding files"
  alreadySeen <- use $ st_knowledge . kd_contents . to ( SM.map ( \ a -> (a, False) ) . seenFiles rootSidPaths )
  missing <- liftM ( SM.keys . SM.filter ( not . snd ) ) $ execStateT ( mapM_ ( addFile' True withRescan ) richPaths ) alreadySeen
  mapM_ ( addMissing sidLocations ) $ sort missing
  cleanLine
  st_storages .= SM.empty
  
  where
    toSidPath :: RichPath -> ( StorageId, FilePath )
    toSidPath rp = ( rp_sid rp, rp_relPath rp )
    seenFiles ::  [(StorageId, FilePath)] -> SM.Map StorageId SeenTree -> SM.Map (StorageId, FilePath) FileId
    seenFiles roots strees = SM.fromList $ concat [ seenFiles' ( SM.findWithDefault ST.empty sid strees ) sid fp | ( sid, fp ) <- roots ]
    seenFiles' stree sid fp = [ ((sid, joinPath $ fp : pcomp ), fid) | (pcomp, fid) <- subFiles ( splitDirectories fp ) stree ]
    addMissing :: ( BufferMonad m, Monad m, Knowledge k, KnownDataClass k ) 
                  => SM.Map StorageId FilePath -> ( StorageId, FilePath ) -> F4LMonad k m  ()
    addMissing sm ( sid, fp ) = case SM.lookup sid sm of
      Nothing -> error $ "Aborted: unable to find storage " ++ sid ++ " within " ++ show sm
      Just sp -> do
        withStorage sid sp
        setPath fp
        addOp Missing

    addFile' :: ( FSMonad m, BufferMonad m, Knowledge k, Show k, KnownDataClass k ) 
                => Bool -> Bool -> RichPath -> StateT ( SM.Map (StorageId, FilePath) ( FileId, Bool ) ) ( F4LMonad k m ) ()
    addFile' withSetPri withRescan rp = do
      isDir <- doesDirectoryExist $ rp_canPath rp
      -- everyN 1000 $ get >>= say Quiet . show . length . show
      if isDir then do
        say Timid $ "Descending into " ++ show ( rp_canPath rp )
        subfps <- lift $ onErrorWarnWithDefault [] $ getCleanDirectoryContents ( rp_canPath rp )
        mapM_ ( \ fp -> addFile' False withRescan ( subRichPath fp rp ) ) subfps
        else do
        -- On error report size -1 to make sure no collision happens
        -- TODO: choose better way?
        s <- lift $ onErrorWarnWithDefault (-1) $ fileSize $ rp_canPath rp
        let k = ( rp_sid rp, rp_relPath rp )
        -- Mark the file as found
        -- TODO: if there is an error below, use a specific state
        modify $ SM.adjust ( _2 .~ True ) k
        seen <- gets $ liftM fst . SM.lookup k
        lift $ onErrorWarn $ do
          ( fid, samePosition ) <- case seen of
            Just fid @ ( FileId l _ ) | l == s && not withRescan -> do
              say Chatty $ "Skipping " ++ show ( rp_canPath rp )
              return ( fid, True )
            mfid -> do
              fid <- fileId $ rp_canPath rp
              lift $ ( \ a -> say Chatty a >> outputLineMsg a ) ( "Adding " ++ rp_canPath rp )
              return ( fid, Just fid == mfid )
          addFile rp samePosition fid
      -- set path priorities
      when withSetPri $ case mpri of
        Nothing -> return ()
        Just pri -> lift $ do
          withStorage ( rp_sid rp ) ( rp_sPath rp )
          setPath ( rp_relPath rp )
          addOp $ WithPriority pri
          addOp MarkPath

==> file4life-1.1.4-source.tgz: file4life-1.1.4/src/F4L/Actions/Files.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Actions.Files (
  addFiles
  , addTestFiles
  , checkedCopyFileCandidates
  , fileCandidates
  , fileId
  , findLocalFileCandidates
  , getCleanDirectoryContents
) where

import F4L
import F4L.Actions.Common
import F4L.Actions.Ops
import F4L.Actions.Storages
import F4L.BufferMonad
import F4L.F4LMonad
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import F4L.Knowledge
import F4L.Op
import F4L.SeenTree as ST hiding ( removeFile )
import F4L.Utils hiding( fileSize )

import Control.DeepSeq
import Control.Lens
import Control.Monad
import Control.Monad.Error( ErrorT(..), throwError )
import Control.Monad.Reader
import Control.Monad.State.Strict
import Data.Either( partitionEithers )
import Data.List( sort )
import qualified Data.Map.Strict as M -- TODO: just use SM
import qualified Data.Map.Strict as SM
import qualified Data.Set as S
import System.FilePath( combine, joinPath, splitDirectories )


fileId :: FSMonad m => FilePath -> ErrorT Error m FileId
-- fileId path = processLengthAndBLFromFile path $ \ l bs -> FileId l ( sha1 bs )
fileId path = actOnLengthAndBLFromFile path act
  where
    act l bs = do
      let s = sha1 bs
      let res = FileId l s
      s `deepseq` return res


-- | Directory contents without . and ..
getCleanDirectoryContents dir = 
  getDirectoryContents dir >>= return . filter ( not . flip elem [ ".", ".." ] ) >>= return . sort


-- | Copy file checking that it has the given FileId. Otherwise, remove the created destination file
checkedCopyFile :: DelFSMonad m => FileId -> FilePath -> FilePath -> ErrorT Error m ()
checkedCopyFile ( FileId l h ) source dest = do
  writeRes <- actOnLengthAndBLFromFile source checkedWrite
  case writeRes of
    Left ( WhileOpening err ) -> do
      throwError err
    Left ( WhileWriting err ) -> do
      removeFile dest
      throwError err
    Right res -> return res
  where
    checkedWrite i bl = runErrorT $ do
      when ( i /= l ) $
        throwError $ WhileOpening $ "Length mismatch copying " ++ source ++ " to " ++ dest ++ ": should be " ++ show l ++ ", is " ++ show i
      ctx <- writeBLToFileAndDo dest newSha1 ( liftM ( liftM return ) updateSha1' ) bl
      let rh = toSha1 ctx
      when ( rh /= h ) $ do
        throwError $ WhileWriting $ "Hash mismatch copying " ++ source ++ " to " ++ dest ++ ": should be " ++ hashToS h ++ ", is " ++ hashToS rh
      return ()


-- | Copy file from one of the given candidates, checking it has the give FileId
checkedCopyFileCandidates :: DelFSMonad m => FileId -> [ FilePath ] -> FilePath -> String -> ErrorT Error ( F4LMonad k m ) ()
checkedCopyFileCandidates fid candidates destPath preMessage =
  foldl mplus ( throwError "No candidate for file" ) $ flip map candidates $ \ sourcePath -> do
    outputLineMsg $ preMessage ++ "Copying " ++ sourcePath ++ " to " ++ destPath
    say Chatty $ "Copying " ++ sourcePath ++ " to " ++ destPath
    checkedCopyFile fid sourcePath destPath


findLocalFileCandidates :: ( FSMonad m, BufferMonad m, KnownDataClass k ) 
                           => FileId -> F4LMonad k m [ FilePath ]
findLocalFileCandidates fid = do
  storages <- storagesForEnv
  fd <- use $ st_knowledge . kd_files . to ( SM.findWithDefault noFileData fid  )
  return $ fileCandidates storages fd


fileCandidates :: SM.Map StorageId [ FilePath ] -> FileData -> [ FilePath ]
fileCandidates storages dat =
  [ combine sseen path
  | ( sid, path ) <- map ( mapSnd shortToPath ) $ S.toList ( dat ^. fd_seen )
  , sseen <- M.findWithDefault [] sid storages
  ]


{-
findFileWithSize :: ( FSMonad m, BufferMonad m ) => Int -> [ FilePath ] -> m ( Maybe FilePath )
findFileWithSize _ [] = return Nothing
findFileWithSize i ( x : xs ) = do
  e <- doesFileExist x
  case e of
    True -> do
      es <- runErrorT $ fileSize x
      case es of
        Right s | s == i -> return ( Just x )
        _ -> findFileWithSize i xs
    False ->
      findFileWithSize i xs
-}


addTestFile :: ( BufferMonad m, Knowledge k ) => Int -> F4LMonad k m ()
addTestFile i = do
  let fid = testFid i
  setFid fid
  setPath $ "test-" ++ show i
  addOp SeenMark
  return ()


addTestFiles :: ( BufferMonad m, Knowledge k ) => Int -> F4LMonad k m ()
addTestFiles i = do
  withStorage "-" "/"
  mapM_ addTestFile [ 1 .. i ]
  return ()


-- | Ops are only added on no error
addFile :: ( FSMonad m, BufferMonad m, Knowledge k, KnownDataClass k ) 
           => RichPath -> Bool -> FileId
           -> ErrorT Error ( F4LMonad k m ) ()
addFile rp samePosition fid = do
  let ( sid, sPath, relPath, path ) = ( rp_sid rp, rp_sPath rp, rp_relPath rp, rp_canPath rp )
  say Chatty $ "Adding " ++ sid ++ ":" ++ sPath ++ ":" ++ relPath
  withExif <- lift $ use st_withExif
  exif_props <- if withExif && not samePosition then lift ( exifProperties path ) else return []
  lift $ do
    withStorage sid sPath
    unless samePosition $ do
      setFid fid
      setPath relPath
      -- TODO: only add new properties
      mapM_ (addOp . attrOp) exif_props
      addOp SeenMark



addFiles :: ( FSMonad m, BufferMonad m, Knowledge k, SeenStoragesClass k, KnownDataClass k, Show k ) => Maybe Priority -> [FilePath] -> F4LMonad k m ()
addFiles mpri paths = do
  st_storages .= SM.empty
  ( withError, richPaths ) <- liftM partitionEithers $ forM paths $ \ p -> do
    res <- runErrorT $ forceGetStorage p
    case res of
      Right res -> return $ Right res
      Left err -> return $ Left ( p, err )
  forM_ withError $ \ ( p, err ) -> do
    say Quiet $ "Warning: error setting StorageId for " ++ p ++ ": " ++ err
  let sidLocations = SM.fromList [ ( rp_sid x, rp_sPath x ) | x <- richPaths ]
  say SayAll $ "richPaths: " ++ show richPaths
  let rootSidPaths = dominants sidPathCmp dominates $ map toSidPath richPaths
  say SayAll $ "rootSidPaths: " ++ show rootSidPaths
  withRescan <- use st_withRescan
  -- accumulate files seen in same position with some size
  say Chatty "Adding files"
  alreadySeen <- use $ st_knowledge . kd_contents . to ( SM.map ( \ a -> (a, False) ) . seenFiles rootSidPaths )
  missing <- liftM ( SM.keys . SM.filter ( not . snd ) ) $ execStateT ( mapM_ ( addFile' True withRescan ) richPaths ) alreadySeen
  mapM_ ( addMissing sidLocations ) $ sort missing
  cleanLine
  st_storages .= SM.empty
  
  where
    toSidPath :: RichPath -> ( StorageId, FilePath )
    toSidPath rp = ( rp_sid rp, rp_relPath rp )
    seenFiles ::  [(StorageId, FilePath)] -> SM.Map StorageId SeenTree -> SM.Map (StorageId, FilePath) FileId
    seenFiles roots strees = SM.fromList $ concat [ seenFiles' ( SM.findWithDefault ST.empty sid strees ) sid fp | ( sid, fp ) <- roots ]
    seenFiles' stree sid fp = [ ((sid, joinPath $ fp : pcomp ), fid) | (pcomp, fid) <- subFiles ( splitDirectories fp ) stree ]
    addMissing :: ( BufferMonad m, Monad m, Knowledge k, KnownDataClass k ) 
                  => SM.Map StorageId FilePath -> ( StorageId, FilePath ) -> F4LMonad k m  ()
    addMissing sm ( sid, fp ) = case SM.lookup sid sm of
      Nothing -> error $ "Aborted: unable to find storage " ++ sid ++ " within " ++ show sm
      Just sp -> do
        withStorage sid sp
        setPath fp
        addOp Missing

    addFile' :: ( FSMonad m, BufferMonad m, Knowledge k, Show k, KnownDataClass k ) 
                => Bool -> Bool -> RichPath -> StateT ( SM.Map (StorageId, FilePath) ( FileId, Bool ) ) ( F4LMonad k m ) ()
    addFile' withSetPri withRescan rp = do
      isDir <- doesDirectoryExist $ rp_canPath rp
      if isDir then do
        say Timid $ "Descending into " ++ show ( rp_canPath rp )
        subfps <- lift $ onErrorWarnWithDefault [] $ getCleanDirectoryContents ( rp_canPath rp )
        mapM_ ( \ fp -> addFile' False withRescan ( subRichPath fp rp ) ) subfps
        else do
        -- On error report size -1 to make sure no collision happens
        -- TODO: choose better way?
        s <- lift $ onErrorWarnWithDefault (-1) $ fileSize $ rp_canPath rp
        let k = ( rp_sid rp, rp_relPath rp )
        -- Mark the file as found
        -- TODO: if there is an error below, use a specific state
        modify $ SM.adjust ( _2 .~ True ) k
        seen <- gets $ liftM fst . SM.lookup k
        lift $ onErrorWarn $ do
          ( fid, samePosition ) <- case seen of
            Just fid @ ( FileId l _ ) | l == s && not withRescan -> do
              say Chatty $ "Skipping " ++ show ( rp_canPath rp )
              return ( fid, True )
            mfid -> do
              fid <- fileId $ rp_canPath rp
              lift $ ( \ a -> say Chatty a >> outputLineMsg a ) ( "Adding " ++ rp_canPath rp )
              return ( fid, Just fid == mfid )
          addFile rp samePosition fid
      -- set path priorities
      when withSetPri $ case mpri of
        Nothing -> return ()
        Just pri -> lift $ do
          withStorage ( rp_sid rp ) ( rp_sPath rp )
          setPath ( rp_relPath rp )
          addOp $ WithPriority pri
          addOp MarkPath
